// This software is Copyright (C) 2005 by Andrew Firth, University of Otago, 
// Dunedin, New Zealand. All rights reserved.

// Programming for making running means of columns 2 and 3 in input logfile.
// Command line input: file name and halfwindow size n (window is 2n + 1).
// Leaves column 1 unchanged.  Takes running mean of column 2 (conservation 
//   scores).  Sqrt of sum of squares for column 3 (stddev of conservation 
//   scores).
// Assumes input file is in nucleotide order.
// To skip gaps (etc) just miss those nucleotides out of the input file.
// Sorts col 2 values in each window and clips off upper clip1 and lower clip2
//   (based on col 2 values) before taking col 2 mean and col 3 sqrt of sum 
//   of squares.

#include <iostream>
#include <cmath>
#include <fstream>
#include <cstdlib>
#include <iomanip>
#include<cstring>
using namespace std;
#define maxlength 50000

// Function declarations
int sort(int nvalues, int values1[maxlength], double values2[maxlength], 
	 double values3[maxlength], int newvalue1, double newvalue2, 
	 double newvalue3);

int main(int argc, char* argv[])
{
  if (6 != argc) {
    cerr << "Aborting; use '" << argv[0] << " logfile " 
	 << "window_half_size clip1 clip2 circular'.\n";
    exit(EXIT_FAILURE);
  }

  int j, k, nlines, col1[maxlength], hfw, nvalues, values1[maxlength];
  int newvalue1, c1, c2, n, place, circular;
  double col2[maxlength], col3[maxlength], sum1, sum2, clip1, clip2;
  double values2[maxlength], values3[maxlength], newvalue2, newvalue3;

  int verbose = 0;

  hfw = atoi(argv[2]);
  circular = atoi(argv[5]);

  clip1 = atof(argv[3]);
  clip2 = atof(argv[4]);
  if (clip1 < 0.||clip2 < 0.||clip1 > 1.||clip2 > 1.) {
    cerr << "\nClipping values must be between 0 and 1.\n";
    exit(EXIT_FAILURE);
  }
  if (verbose) {
    cout << "\nClipping values: " << int(0.5 + clip1 * float(2*hfw+1)) 
	 << " " << int(0.5 + clip2 * float(2*hfw+1)) << "\n";
  }

  // Initialize arrays
  for (k = 0; k < maxlength; ++k) {
    col1[k] = 0;
    col2[k] = col3[k] = 0.;
  }

  // Open log file.
  ifstream logfile(argv[1]); 
  if (!logfile) {
    cerr << "Aborting: can't find log file '" << argv[1] << "'.\n";
    exit(EXIT_FAILURE);
  }

  // Number of lines in log file
  nlines = -1;
  while (logfile.ignore(1000, '\n')) {
    ++nlines;
  }
  logfile.clear();
  logfile.seekg(0);
  if (verbose) {
    cout << "There are " << nlines << " data lines.\n\n";
  }

  // Read in logfile
  for (k = 0; k < nlines; ++k) {
    logfile >> col1[k];
    logfile >> col2[k];
    logfile >> col3[k];
    logfile.ignore(1000, '\n');
  }

  // Running mean
  nvalues = 0;
  for (k = 0; k < hfw; ++k) { // read-in first hfw values
    nvalues = sort(nvalues, values1, values2, values3, k, col2[k], col3[k]);
  }
  if (circular) {  // also read-in end hfw values
    for (k = 0; k <= hfw; ++k) {
      nvalues = sort(nvalues, values1, values2, values3, nlines-1-k, 
		     col2[nlines-1-k], col3[nlines-1-k]);
    }
  }
  for (k = 0; k < nlines; ++k) {
    if (k > hfw || circular) {  // take out nt number (k-hfw-1+nlines)%nlines
      for (j = 0; j < nvalues; ++j) {  
	if ((k-hfw-1+nlines)%nlines == values1[j]) {  
	  place = j;
	  break;
	}
      }
      for (j = place; j < nvalues; ++j) {     
	values1[j] = values1[j+1];
	values2[j] = values2[j+1];
	values3[j] = values3[j+1];
      }
      nvalues--;
    }
    if (k+hfw < nlines || circular) {  // add in next value
      nvalues = sort(nvalues, values1, values2, values3, (k+hfw+nlines)%nlines,
		     col2[(k+hfw+nlines)%nlines], col3[(k+hfw+nlines)%nlines]);
    }

    c1 = int(0.5 + clip1 * nvalues); // upper cutoff, remove the last c1 nt
    c2 = int(0.5 + clip2 * nvalues); // lower cutoff, remove the first c2 nt

    sum1 = sum2 = 0.;
    n = 0;
    for (j = c2; j < nvalues - c1; ++j) { // clipped sum/sum of squares
      sum1 += values2[j];
      sum2 += values3[j] * values3[j];
      n += 1;
    }

    cout << col1[k] << " " << sum1/float(n) << " " << sqrt(sum2)/float(n) 
	 << "\n";
    //    for (j = 0; j < nvalues; ++j) {cout << values2[j] << " ";}
    //    cout << "\n";
  }

}


int sort(int nvalues, int values1[maxlength], double values2[maxlength], 
	 double values3[maxlength], int newvalue1, double newvalue2, 
	 double newvalue3)

{
  int i, m1, m2, mm, place, test;

  if (nvalues <= 0) {
    values1[0] = newvalue1;
    values2[0] = newvalue2;
    values3[0] = newvalue3;
    return(1);
  }

  if (newvalue2 <= values2[0]) {
    place = 0;
  } else if (newvalue2 > values2[nvalues-1]) {
    place = nvalues;
  } else {
    m1 = 0;
    m2 = nvalues-1;
    test = m2-m1;
    while (test > 1) {
      mm = (m1+m2)/2;
      if (newvalue2 <= values2[mm]) {
	m2 = mm;
      } else {
	m1 = mm;
      }
      test = m2-m1;
    }
    place = m2;
  }

  for (i = nvalues-1; i >= place; --i) {
    values1[i+1] = values1[i];
    values2[i+1] = values2[i];
    values3[i+1] = values3[i];
  }
  values1[place] = newvalue1;
  values2[place] = newvalue2;
  values3[place] = newvalue3;

  return(nvalues+1);
}
